type Element = f64;

use std::ops::{Index, IndexMut};

#[derive(Debug, PartialEq)]
pub struct Matrix {
    pub n: usize,
    values: Vec<Element>,
}

impl Matrix {
    pub fn new(n: usize, values: Vec<Element>) -> Option<Self> {
        if n.pow(2) == values.len() {
            Some(Matrix { n, values })
        } else {
            None
        }
    }

    pub fn new_lattice(noise_bits: usize, ciphertexts: Vec<Element>) -> Option<Self> {
        let n = ciphertexts.len();
        let mut values = Vec::with_capacity(n * n);

        // First row: [2^noise_bits, ciphertexts[0], ciphertexts[1], ..., ciphertexts[t]]
        values.push((2.0f64).powi(noise_bits as i32));
        values.extend_from_slice(&ciphertexts[1..]);

        // -x0 on diagonal, 0 everywhere else
        let x0 = ciphertexts[0];
        for i in 0..n - 1 {
            let mut row = vec![0.0; n];
            row[i] = -x0;
            values.extend(row);
        }

        Matrix::new(n, values)
    }
}

impl Index<(usize, usize)> for Matrix {
    type Output = f64;
    fn index(&self, index: (usize, usize)) -> &Self::Output {
        &self.values[(self.n * index.0) + index.1]
    }
}

impl IndexMut<(usize, usize)> for Matrix {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        &mut self.values[(self.n * index.0) + index.1]
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    #[test]
    fn simple_matrix() {
        assert_eq!(
            Matrix {
                n: 2,
                values: Vec::from([1.0, 2.0, 3.0, 4.0])
            },
            Matrix::new(2, Vec::from([1.0, 2.0, 3.0, 4.0])).unwrap()
        );
        assert!(Matrix::new(3, Vec::from([1.0, 2.0])).is_none());
    }
    #[test]
    fn indexes() {
        let mut m = Matrix::new(2, vec![1.0, 2.0, 3.0, 4.0]).unwrap();
        assert_eq!(m[(0, 0)], 1.0);
        assert_eq!(m[(1, 0)], 3.0);

        m[(1, 0)] = 5.0;
        assert_eq!(m[(1, 0)], 5.0);
    }
    #[test]
    fn test_new_lattice() {
        let ciphertexts = vec![5.0, 8.0, 12.0]; // x0=5.0, x1=8.0, x2=12.0
        let noise_bits = 2;

        // Expected matrix (3x3):
        // [ 2^2,   x1,   x2 ] = [4.0, 8.0, 12.0]
        // [ -x0,    0,    0 ] = [-5.0, 0.0, 0.0]
        // [   0,  -x0,    0 ] = [0.0, -5.0, 0.0]
        let expected_values = vec![4.0, 8.0, 12.0, -5.0, 0.0, 0.0, 0.0, -5.0, 0.0];

        let lattice = Matrix::new_lattice(noise_bits, ciphertexts).unwrap();
        assert_eq!(lattice.n, 3);
        assert_eq!(lattice.values, expected_values);
    }
}
